/* DICElib (DIstributed CAVE Engine library) 
 * Copyright (c) 2001 Bruno Barberi Gnecco <brunobg@lsi.usp.br>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "_dicelib.h"
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#ifndef WIN32
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
# ifdef USE_LTDL
#  include <ltdl.h>
# endif
#else
#define VERSION "Windows Port"
#include <winsock.h>
#endif

#ifdef USE_MCHECK
#include <mcheck.h>
#endif

static struct {
	int 		libinit;
	int		datainit;
} _int_init = { 0, 0 };

#ifndef WIN32
pid_t pid = 0;
#endif
char *defaultpath;
char **environ;
static int dice_parse_dicerc ( void );

struct _dice_data dice_data;
#define NO_FUNC ""

static void dice_datainit ( void ) {
	dice_data.verbose = 3;
	dice_data.port = 6748;
	dice_data.max_connections = 5;
	dice_data.error = stderr;
	dice_data.handle = NULL;
	dice_data.color[0] = NULL;
	dice_data.color[1] = NULL;
	dice_data.clientnum = -1;
	dice_data.clients = NULL;
	dice_data.id = -1;
	dice_data.controlname = NULL;
	dice_data.original_only = 0;
	dice_data.spawn = RSH;

	/* todo: find the multicast address from ioctl or something */
	inet_aton("239.0.0.1", &dice_data.mcastaddr);
	dice_data.mcastport = dice_data.port+1;
	dice_data.mcastTTL = 1;
	dice_data.mcastsock = -1;
}

#define FUNCTION 	"DICE_init"
/**
  \brief Inits the library.
 
  This function must be called before any other DICE function.
  Recognized arguments are:
  
  - \e --dice-client-num \e [number]: expect \c number clients. If used with
  a rc file, spawn \c number clients.
  - \e --dice-client-only: do not start client.
  - \e --dice-control-is \e [name]: use \c name as control server name.
  - \e --dice-control-only: do not start control.
  - \e --dice-environment \e [env]: set the \c \env environment in the clients.
  Useful for X11 remote display.
  - \e --dice-no-rc: don't read rc file.
  - \e --dice-spawn [type]: type of spawning. /c Type is either \c none, \c rsh,
  \c ssh or \c bproc. The default is \c rsh.
  - \e --dice-original: only original node can update.
  - \e --dice-output \e [file]: output error messages to \c file. The default is
  stderr.
  - \e --dice-port \e [port]: use \c port as port. The default is 6748.
  - \e --dice-verbose \e [level]: set verbose level to \c level.

  DICE_init "cleans" argc and argv, taking out all arguments that are used.

  \param argc pointer number of arguments.
  \param argv pointer to the arguments.
  \retval 0 OK
  \retval -1 error.
*/
int DICE_init ( int *argc, char **argv ) {
	int			i;
	struct {		int norc;
	} flags = { 0 };
	enum { CONTROL, CLIENT } type = CONTROL;
	char			*ourargv[25]; /*todo: dynamic */
	char			dir[256], *tmp;
	int			ourargc, used_args = 0;

	if ( _int_init.libinit == 0 )
		_int_init.libinit = 1;
	else
		return -1;

#ifdef USE_MCHECK
	mtrace(); /* memory leaks */
#endif

//	putenv("DISPLAY=:0.0");
	/* init defaults */
	if (_int_init.datainit == 0 ) {
		dice_datainit();
		_int_init.datainit = 1;
	}
	_dice_debug(3, fprintf(dice_data.error, FUNCTION "()\n");)

#ifdef USE_LTDL
	if ( lt_dlinit() ) {
		_dice_debug(1, fprintf(dice_data.error, FUNCTION "Error in ltdl: %s", lt_dlerror());)
		return -11;
	}
	dice_data.handle = lt_dlopen(NULL);
	if ( !dice_data.handle ) {
		_dice_debug(1, fprintf(dice_data.error, FUNCTION " %s\n", lt_dlerror());)
		return -1;
	}
#endif

	/* parse arguments; and get a copy for spawning */
	/* ourargv[0]: machine name.
	   ourargv[1]: program name.
	   ourargv[2]: --dice-control-is
	   ourargv[3]: control address
	*/
	/* set program name */
	tmp = strrchr(argv[0], '/');
	if ( !tmp )
		tmp = argv[0];
	else
		tmp++;
#ifndef WIN32
	getcwd(dir, 256);
#else
	dir[0] = '\0';
#endif
	defaultpath = malloc(strlen(dir)+strlen(tmp)+2);
	strcpy(defaultpath, dir);
	strcat(defaultpath, "/");
	strcat(defaultpath, tmp);
	ourargv[2] = strdup("--dice-control-is");
	ourargc = 4;

	for ( i = 1; i < *argc; ) {
		if ( strcmp(argv[i], "--dice-client-num" ) == 0 ) {
			dice_data.clientnum = atoi(argv[i+1]);
			_dice_debug(3, dice_printf(FUNCTION, "using %d clients.\n",
					dice_data.clientnum);)
			if (i+2 <= *argc)
				memmove(&argv[i], &argv[i+2], sizeof(char *)*(*argc-i));
		}
		else if ( strcmp(argv[i], "--dice-client-only" ) == 0 ) {
			_dice_debug(3, dice_printf(FUNCTION, "We're CLIENT\n");)
			type = CLIENT;
			used_args = 1;
		}
		else if ( strcmp(argv[i], "--dice-control-is" ) == 0 ) {
			_dice_debug(3, dice_printf(FUNCTION, "control is: %s\n",
					argv[i+1]);)
			dice_data.controlname = argv[i+1];
			used_args = 2;
			type = CLIENT;
		}
		else if ( strcmp(argv[i], "--dice-control-only" ) == 0 ) {
			_dice_debug(3, dice_printf(FUNCTION, "We're CONTROL\n");)
			type = CONTROL;
			(*argc)--;
			if (i+1 <= *argc)
				memmove(&argv[i], &argv[i+1], sizeof(char *)*(*argc-i));
		}
		else if ( strcmp(argv[i], "--dice-environment" ) == 0 ) {
			_dice_debug(3, dice_printf(FUNCTION, "environment: %s\n",
					argv[i+1]);)
			putenv(argv[i+1]);
			used_args = 2;
		}
#if 0
		else if ( strcmp(argv[i], "--dice-function" ) == 0 ) {
			_dice_debug(3, dice_printf(FUNCTION, "function is: %s\n",
					argv[i+1]);)
			if ( *argv[i+1] != '!' ) {
				dice_data.function = dlsym(dice_data.handle, argv[i+1]);
				if ( !dice_data.function ) {
					_dice_debug(2, dice_printf(FUNCTION, "function %s not found: %s\n",
						argv[i+1], dlerror());)
				}
			}
			used_args = 2;
		}
#endif
		else if ( strcmp(argv[i], "--dice-no-rc" ) == 0 ) {
			_dice_debug(3, dice_printf(FUNCTION, "don't read rc.\n");)
			flags.norc = 1;
			used_args = 1;
		}
		else if ( strcmp(argv[i], "--dice-spawn" ) == 0 ) {
			if ( strcmp(argv[i+1], "none") == 0 )
				dice_data.spawn = NONE;
			else if ( strcmp(argv[i+1], "rsh") == 0 )
				dice_data.spawn = RSH;
			else if ( strcmp(argv[i+1], "ssh") == 0 )
				dice_data.spawn = SSH;
			else if ( strcmp(argv[i+1], "bproc") == 0 )
				dice_data.spawn = BPROC;
			else  {
				_dice_debug(1, dice_printf(FUNCTION, "Invalid spawn option: %s\n", argv[i+1]);)
				i++;
				continue;
			}
			_dice_debug(3, dice_printf(FUNCTION, "Spawn type: %s.\n", argv[i+1]);)
			*argc -= 2;
			if (i+2 <= *argc)
				memmove(&argv[i], &argv[i+2], sizeof(char **)*(*argc-i));
		}
#if 0
		else if ( strcmp(argv[i], "--dice-original" ) == 0 ) {
			_dice_debug(3, dice_printf(FUNCTION, "original only.\n");)
			dice_data.original = 1;
			used_args = 1;
		}
#endif
		else if ( strcmp(argv[i], "--dice-output" ) == 0 ) {
			FILE		*out;
			if ( (out = fopen(argv[i+1], "w")) != NULL ) {
				dice_data.error = out;
				_dice_debug(3, dice_printf(FUNCTION, 
					"output will be sent to %s\n",
					argv[i+1]);)
			}
			used_args = 2;
		}
		else if ( strcmp(argv[i], "--dice-port" ) == 0 ) {
			dice_data.port = atoi(argv[i+1]);
			_dice_debug(3, dice_printf(FUNCTION, "port set to %d\n",
					dice_data.port);)
			used_args = 2;
		}
		else if ( strcmp(argv[i], "--dice-verbose" ) == 0 ) {
			dice_data.verbose = atoi(argv[i+1]);
			_dice_debug(1, dice_printf(FUNCTION, "verbose set to %d\n",
					dice_data.verbose);)
			used_args = 2;
		}
		else if ( strncmp(argv[i], "--dice", 6) == 0 ) {
			_dice_debug(2, dice_printf(FUNCTION, "invalid argument %s\n",
					argv[i]);)
			i++;
		}
		else {
			i++;
		}
		if (used_args) {
			memcpy(&ourargv[ourargc], &argv[i], used_args*sizeof(char *));
			ourargc += used_args;
			*argc -= used_args;
			if (i+used_args <= *argc)
				memmove(&argv[i], &argv[i+used_args], sizeof(char *)*(*argc-i));
			used_args = 0;
		}
	}
	ourargv[ourargc] = NULL;


	/* other initializations */
	if ( dice_type_init_defaults() == -1 ) {
		_dice_debug(1, dice_printf(FUNCTION, "init_types\n");)
		return -1;
	}
	if ( dice_sync_init() == -1 ) {
		_dice_debug(1, dice_printf(FUNCTION, "sync_init\n");)
		return -1;
	}
#ifdef USE_ASYNC
	if ( dice_async_init() == -1 ) {
		_dice_debug(1, dice_printf(FUNCTION, "async_init\n");)
		return -1;
	}
#endif

	/* try to read data file. SHOULD BE DONE ONLY IN CONTROL, i.e., after fork!! */
	if ( flags.norc == 0 && type == CONTROL ) {
		if (dice_parse_dicerc())
			return -1;
	}

	if ( type == CONTROL ) { /* child or server only*/
		dice_data.color[0] = strdup("[0;31m");
		dice_data.color[1] = strdup("[0m");
		dice_data.controlname = (char *)malloc(127);
		if ( !dice_data.controlname ) {
			_dice_debug(1, dice_printf(FUNCTION, "%s\n", strerror(errno));)
			return -1;
		}
		gethostname(dice_data.controlname, 127);
		_dice_debug(3, dice_printf(FUNCTION, "I'm the server %s\n",
				dice_data.controlname);)
		_exit(dice_control_loop(ourargc, ourargv));
	}
	else if ( type == CLIENT ) {
		dice_data.color[0] = NULL;
		dice_data.color[1] = NULL;

		_dice_debug(3, dice_printf(FUNCTION, "I'm the client\n");)
		if ( !dice_data.controlname ) { /* assume localhost */
			dice_data.controlname = (char *)malloc(127);
			if ( !dice_data.controlname ) {
				_dice_debug(1, dice_printf(FUNCTION, "%s\n", strerror(errno));)
				return -1;
			}
			gethostname(dice_data.controlname, 127);
		}
		_dice_debug(3, dice_printf(FUNCTION, "Control is %s\n", dice_data.controlname);)
	}

	dice_data.clientnum = -1; /* to know we are clients */
	/* client */
	return dice_init_client();
}

#undef FUNCTION
#define FUNCTION 		"DICE_close"
/** 
  \brief Close library.

  This function must be called when you want to close the DICE application.
*/
void DICE_close ( void ) {
	dice_packet		bye;
	_dice_debug(3, dice_printf(FUNCTION, "()\n");)

	if ( _int_init.libinit == 0 )
		return;

	bye.packtype = GOODBYE;
	dice_packet_send(&bye, dice_data.controlsock);

#ifdef USE_ASYNC
	dice_async_end();
#endif
	dice_sync_end();
	dice_type_free_all();
#ifdef USE_LTDL
	lt_dlexit();
#endif
//	dice_free(dice_data.controlname);
	dice_free(dice_data.color[0]);
	dice_free(dice_data.color[1]);
#ifndef WIN32
	close(dice_data.controlsock);
	if (dice_data.mcastsock != -1) {
		close(dice_data.mcastsock);	
	}
#else
	closesocket(dice_data.controlsock);
	if (dice_data.mcastsock != -1) {
		closesocket(dice_data.mcastsock);
	}
#endif
}

#undef FUNCTION
#define FUNCTION		"DICE_set_attribute"
/**
  \brief Sets an attribute value.
  See enum for list
  \param t Attribute type.
  \param value Attribute value, cast to void *.
  \sa DICE_getAttribute.
  \retval 0 OK
  \retval -1 error.
*/
int DICE_set_attribute ( DICE_AttributeType t, void *value ) {
	struct in_addr addr;
	_dice_debug(3, dice_printf(FUNCTION, "(%d, %p)\n", t, value);)

	if (_int_init.datainit == 0 ) {
		dice_datainit();
		_int_init.datainit = 1;
	}
	switch ( t ) {
	  case LIBVERSION:
		_dice_debug(2, dice_printf(FUNCTION, " VERSION is read-only.\n");)
		return 0;
	  case VERBOSE:
		if ( (int)value < 0 || (int)value > 3 ) {
			_dice_debug(2, dice_printf(FUNCTION, " VERBOSE out of bonds: %d\n", (int)value);)
			return -1;
		}
		dice_data.verbose = (int)value;
		return 0;
	  case PORT:
		if ( (int)value < 1024 || (int)value > 65535 ) {
			_dice_debug(2, dice_printf(FUNCTION, " PORT out of bonds: %d\n", (int)value);)
			return -1;
		}
		dice_data.port = (int)value;
		return 0;
	  case MAX_CONNECTIONS:
		if ( (int)value <= 0 && (int)value < 5 ) { /* BSD limit it to 5 */
			_dice_debug(2, dice_printf(FUNCTION, " MAX_CONNECTIONS out of bonds: %d\n", (int)value);)
			return -1;
		}
		dice_data.max_connections = (int)value;
		return 0;
	  case OUTPUT:
		if ( value == NULL ) {
			_dice_debug(2, dice_printf(FUNCTION, " NULL file\n");)
			return -1;
		}
		dice_data.error = (FILE *)value;
		return 0;
	  case COLOR:
		return 0;
	  case MULTICAST_ADDRESS: 
		if ( value == NULL ) {
			_dice_debug(2, dice_printf(FUNCTION, " NULL address\n");)
			return -1;
		}
		if (!inet_aton((char *)value, &addr)) {
			_dice_debug(1, dice_printf(FUNCTION, " Couldn't convert address %s!\n");)
		}
		dice_data.mcastaddr = addr;
		return 0;
	  case MULTICAST_PORT: 
		if ( (int)value < 1024 || (int)value > 65535 ) {
			_dice_debug(2, dice_printf(FUNCTION, " MULTICAST_PORT out of bonds: %d\n", (int)value);)
			return -1;
		}
		dice_data.mcastport = (int)value;
		return 0;
	  case MULTICAST_TTL:
		if ( (unsigned int)value < 0 || (unsigned int)value > 255 ) {
			_dice_debug(2, dice_printf(FUNCTION, " MULTICAST_TTL out of bonds: %d\n", (int)value);)
			return -1;
		}
		dice_data.mcastTTL = (unsigned int)value;
		return 0;
	  default:
		_dice_debug(2, dice_printf(FUNCTION, " No such attribute %d\n", t);)
		return -1;
	}
}

#undef FUNCTION
#define FUNCTION		"DICE_get_attribute"
/**
  \brief Gets an attribute value.

  \param t Attribute type.
  \sa DICE_setAttribute.
  \return The attribute value, cast to (void *).
*/
void *DICE_get_attribute ( DICE_AttributeType t ) {
	_dice_debug(3, dice_printf(FUNCTION, "(%d)\n", t);)

	if (_int_init.datainit == 0 ) {
		dice_datainit();
		_int_init.datainit = 1;
	}
	switch ( t ) {
	  case LIBVERSION:
		return (void *)VERSION;
	  case VERBOSE:
		return (void *)dice_data.verbose;
	  case PORT:
		return (void *)dice_data.port;
	  case MAX_CONNECTIONS:
		return (void *)dice_data.max_connections;
	  case OUTPUT:
		return (void *)dice_data.color;
	  case COLOR:
		return (void *)dice_data.color;
	  case MULTICAST_ADDRESS: 
		return (void *)&dice_data.mcastaddr;
	  case MULTICAST_PORT: 
		return (void *)dice_data.mcastport;
	  case MULTICAST_TTL:
		return (void *)((unsigned int)dice_data.mcastTTL);
	  default:
	    _dice_debug(2, dice_printf(FUNCTION, " No such attribute %d\n", t);)
	}
	return NULL;
}

/**
  \brief Our own printf.
  Formats a string to be printed with colors (using dice_data.color[]),
  sends to dice_data.error, etc.
  \param function The function name, may be NULL.
  \param format A printf-like format string.
  \param ... Other arguments.
*/
void dice_printf ( char *function, char *format, ... ) {
	va_list			argp;

	if ( dice_data.color[0] )
		fputs(dice_data.color[0], dice_data.error);
	if ( function ) {
		fputs(function, dice_data.error);
		fputs(": ", dice_data.error);
	}
	va_start(argp, format);
	vfprintf(dice_data.error, format, argp);
	va_end(argp);
	if ( dice_data.color[1] )
		fputs(dice_data.color[1], dice_data.error);
}

#undef FUNCTION
#define FUNCTION		"dice_parse_dicerc"
/**
  \brief Parses the dicerc file.

  \retval 0 OK
  \retval -1 error.
*/
static int dice_parse_dicerc ( void ) {
	FILE *fp;
	char buf[256], *t;
	int lines = 0;

	/* try to open file */
	fp = fopen("dicerc", "r");
	if (!fp)
		fp = fopen("~/dicerc", "r");
	if (!fp)
		fp = fopen("/etc/dicerc", "r");

	if (!fp) {
		_dice_debug(1, dice_printf(FUNCTION, "Could not find dicerc\n");)
		return -1;
	}

	/* parse */
	/* number of clients */
	while (!feof(fp)) {
		fgets(buf, 256, fp);
		t = strtok(buf, " \t\r\n");
		if (t && *t != '#')	/* filter comments */
			lines++;
	}
	rewind(fp);
	if (dice_data.clientnum <= 0 || dice_data.clientnum > lines)
		dice_data.clientnum = lines;

	dice_data.clients = malloc(dice_data.clientnum*sizeof(struct _clients));
	if (!dice_data.clients) {
		_dice_debug(1, dice_printf(FUNCTION, "Malloc clients.\n");)
		return -1;
	}

	/* get client names */
	lines = 0;
	while (!feof(fp) && lines < dice_data.clientnum) {
		fgets(buf, 256, fp);
		t = strtok(buf, " \t\r\n");
		if (t && *t != '#' ) {
			dice_data.clients[lines].address = strdup(t);
			if (!dice_data.clients[lines].address) {
				_dice_debug(1, dice_printf(FUNCTION, "strdup\n");)
				return -1;
			}
		}
		else
			continue;

		if (!(t = strtok(NULL, " \t\r\n"))) {
			dice_data.clients[lines].path = NULL;
			dice_data.clients[lines].function = NO_FUNC;
			continue;
		}
		
		/* get path */
		if (*t == '%')
			dice_data.clients[lines].path = NULL;
		else {
			dice_data.clients[lines].path = strdup(t);
			if (!dice_data.clients[lines].path) {
				_dice_debug(1, dice_printf(FUNCTION,
							"strdup 2\n");)
					return -1;
			}
		}
		/* get functions */
		if ((t = strtok(NULL, " \t\r\n"))) {
			dice_data.clients[lines].function = strdup(t);
			if (!dice_data.clients[lines].function) {
				_dice_debug(1, dice_printf(FUNCTION,
							"strdup 3\n");)
				return -1;
			}
		}
		else
			dice_data.clients[lines].function = NO_FUNC;

		dice_printf(FUNCTION, "[%s %s %s]\n",
			    dice_data.clients[lines].address,
			    dice_data.clients[lines].path,
			    dice_data.clients[lines].function);
		lines++;
	}
	fclose(fp);
	return 0;
}


/* The following is only used for documentation, and can be disregarded. */
/*! \mainpage DICE API doxygen documentation
 \section Development

  etc	

 */
