/* DICElib (DIstributed CAVE Engine library) 
 * Copyright (c) 2001 Bruno Barberi Gnecco <brunobg@lsi.usp.br>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "_dicelib.h"
#include "hash.h"
#include <string.h>
#include <signal.h>
#include <fcntl.h>


static HashTable dice_sync_var;
extern int packetfrom;

/*
 * List to store operations
 */
struct dice_sync_var_item {
	struct dice_sync		*ds;
	struct dice_sync_var_item	*next;
};

static struct dice_sync_var_item *sync_head, *sync_last;

#define dice_sync_add_to_list(ds)	\
	item = (struct dice_sync_var_item *)malloc(sizeof(struct dice_sync_var_item)); \
	if ( !item ) { \
		_dice_debug(1, dice_printf(FUNCTION, "NULL malloc 2\n");) \
		free(ds); \
		return -1; \
	} \
	item->ds = ds; \
	item->next = NULL; \
	if ( !sync_head ) \
		sync_head = item; \
	if ( sync_last ) \
		sync_last->next = item; \
	sync_last = item;

/**
  \brief Synchronous variable states.
*/
enum dice_variable_state {
	UPTODATE,				/**< Up to date. */
	CHANGED, 				/**< Changed by this node. */
	CREATED, 				/**< Created after last DICE_sync() */
	DELETED					/**< Deleted after last DICE_sync() */
};

/**
  \brief Synchronous variable type.
*/
struct dice_sync {
	char			*name;		/**< Variable name. */
	char			*type;		/**< Variable type. */
	void			*value;		/**< Variable value. */
	void			*nextvalue;	/**< Value for next update */
	enum dice_variable_state	state;	/**< Variable state */
	
	/* server only */
	int			owner;		/**< Socket of variable creator. */
	int			changedby;	/**< Socket of last client that changed. */
};

/**
  \brief Frees a sync variable
  \param data The data to free.
*/
void dice_sync_free ( void *data ) {
	struct dice_sync 	*ds;
	if ( !data )
		return;

	ds = data;
	dice_free(ds->name);
	DICE_type_free(ds->type, ds->value);
//	DICE_type_free(ds->type, ds->nextvalue);
	dice_free(ds->type);
	free(ds);
}

/**
  \brief Initializes synchronous variables 
*/
int dice_sync_init ( void ) {
	if ( hash_init(&dice_sync_var, 128, NULL) == -1 )
		return -1;
	sync_head = NULL;
	sync_last = NULL;
	return 0;
}

/**
 \brief Frees synchronous variables 
 */
void dice_sync_end ( void ) {
	struct dice_sync_var_item 		*item, *tmp;
	if ( sync_head )
		for ( item = sync_head; item != NULL; ) {
			if ( item->ds->state == CREATED )
				free(item->ds);
			tmp = item->next;
			free(item);
			item = tmp;
		}
	hash_free(&dice_sync_var, dice_sync_free);
}

#define FUNCTION		"dice_sync_update_send"
/**
  \brief Send all updates.
 
  Long description
*/
void dice_sync_update_send ( void ) {
	struct dice_sync_var_item	*item, *tmp;
	dice_packet			packet;

	_dice_debug(3, dice_printf(FUNCTION, "()\n");)

	for ( item = sync_head; item != NULL; item = item->next ) {
		switch ( item->ds->state ) {
		  case CREATED:
		  case CHANGED:
			packet.packtype = VARIABLE_SYNC;
			packet.variable.name = item->ds->name;
			packet.variable.type = item->ds->type;
			packet.variable.value = (item->ds->state == CREATED ?
					item->ds->value : item->ds->nextvalue );
			if ( dice_is_server() ) {
				dice_broadcast(&packet, -1);
				item->ds->state = UPTODATE;
			}
			else
				dice_packet_send(&packet, dice_data.controlsock);
			break;
		  case DELETED:
			packet.packtype = VARIABLE_FREE_SYNC;
			packet.variable_free.name = item->ds->name;
			if ( dice_is_server() )
				dice_broadcast(&packet, -1);
			else
				dice_packet_send(&packet, dice_data.controlsock);
			break;
		  case UPTODATE:
			break;
		}
	}

	/* garbage collection */
	for ( item = sync_head; item != NULL; ) {
		if ( item->ds->state == CREATED ) {
			dice_free(item->ds->name);
			dice_free(item->ds->type);
			free(item->ds);
		}
		else {
			item->ds->nextvalue = NULL;
			if ( dice_is_server() )
				item->ds->changedby = -1;
		}
		tmp = item;
		item = item->next;
		free(tmp);
	}
	sync_head = sync_last = NULL;
}

#undef FUNCTION
#define FUNCTION		"dice_sync_new"
/**
  \brief Internal: Creates a new synchronized variable.
 
  Long description

  \param name The variable name.
  \param type The variable type. Must be previously registered.
  \param value The variable value, converted to void *.
  \sa DICE_type_register, dice_sync_delete.
  \retval 0 OK.
  \retval -1 error.
  \retval -2 Variable already exists.
*/
int dice_sync_new ( char *name, char *type, void *value, 
				enum dice_variable_state state ) {
	struct dice_sync		*ds;
	struct dice_sync_var_item	*item;
	int				exists = 0;

	_dice_debug(3, dice_printf(FUNCTION, "(%s,%s,%p)\n",
			name, type, value);)

	if ( !name ) {
		_dice_debug(1, dice_printf(FUNCTION, "Invalid name\n");)
		return -1;
	}
	
	if ( !dice_type_is_valid(type) ) {
#if 0
		if ( DICE_type_register(type) == -1 ) {
			_dice_debug(1, dice_printf(FUNCTION, "Couldn't register type.\n");)
			return -1;
		}
#endif
		_dice_debug(1, dice_printf(FUNCTION, "Invalid type %s.\n", type);)
		return -1;
	}
	for ( item = sync_head; item != NULL; item = item->next )
		if ( strcmp(item->ds->name, name) == 0 ) {
			exists = 1;
			break;
		}

	if ( (ds = hash_data(&dice_sync_var, name)) != NULL || exists ) {
		_dice_debug(2, dice_printf(FUNCTION, "Variable exists\n");)
		return -2;
	}
	
	ds = (struct dice_sync *)malloc(sizeof(struct dice_sync));
	if ( !ds ) {
		_dice_debug(1, dice_printf(FUNCTION, "NULL malloc\n");)
		return -1;
	}

	ds->name = strdup(name);
	if ( !ds->name ) {
		_dice_debug(1, dice_printf(FUNCTION, "NULL strdup\n");)
		return -1;
	}
	ds->type = strdup(type);
	if ( !ds->type ) {
		_dice_debug(1, dice_printf(FUNCTION, "NULL strdup2\n");)
		return -1;
	}
	ds->value = value;
	ds->nextvalue = NULL;

	if ( state == CREATED )	{
		ds->state = CREATED;
		dice_sync_add_to_list(ds);
		if ( dice_is_server() ) {
			ds->owner = ds->changedby = packetfrom;
			hash_insert(&dice_sync_var, ds->name, ds);
		}
	}
	else if ( state == UPTODATE ) {
		ds->nextvalue = NULL;
		ds->state = UPTODATE;
		hash_insert(&dice_sync_var, ds->name, ds);
	} else {
		_dice_debug(1, dice_printf(FUNCTION, "Wrong state!\n");)
		/* clean */
		return -1;
	}

	return 0;
}

#undef FUNCTION
#define FUNCTION		"dice_sync_delete"
/**
  \brief Internal: Deletes an existing synchronized variable.
 
  Deletes an existing synchronized variable. Takes care of memory deallocation.

  \param name The variable name.
  \sa DICE_sync_new.
*/
void dice_sync_delete ( char *name ) {
	struct dice_sync		*ds;

	_dice_debug(3, dice_printf(FUNCTION, "(%s)\n", name);)

	if ( !name ) {
		_dice_debug(2, dice_printf(FUNCTION, "NULL name.\n");)
		return;
	}
	ds = hash_del(&dice_sync_var, name);
	if ( !ds ) {
		_dice_debug(2, dice_printf(FUNCTION, "variable not found.\n");)
		return;
	}

	dice_sync_free((void *)ds);
}

#undef FUNCTION
#define FUNCTION		"dice_sync_update"
/**
  \brief Internal: updates an existing synchronized variable.
 
  Deletes an existing synchronized variable. Takes care of memory deallocation.

  \param name The variable name.
  \sa DICE_sync_new.
*/
void dice_sync_update ( char *name, char *type, void *value ) {
	struct dice_sync		*ds;

	_dice_debug(3, dice_printf(FUNCTION, "(%s)\n", name);)

	if ( !name || !dice_type_is_valid(type)) {
		_dice_debug(2, dice_printf(FUNCTION, "NULL name or invalid type.\n");)
		return;
	}
	ds = (struct dice_sync *)hash_data(&dice_sync_var, name);
	if ( !ds ) {
		dice_sync_new(name, type, value, UPTODATE);
		return;
	}

	DICE_type_free(type, ds->value);
	if ( ds->nextvalue )
		DICE_type_free(type, ds->nextvalue); 
	ds->nextvalue = NULL;
	ds->value = value;
	ds->state = UPTODATE;
	return;
}

/*
 * Public functions
 */

#undef FUNCTION
#define FUNCTION		"DICE_sync_new"
/**
  \brief Creates a new synchronized variable.
 
  Long description

  \param name The variable name.
  \param type The variable type. Must be previously registered.
  \param value The variable value, converted to void *.
  \sa DICE_type_register, dice_sync_delete.
  \retval 0 OK.
  \retval -1 error.
  \retval -2 Variable already exists.
  \retval -3 You do not have permission to create variables.
*/
int DICE_sync_new ( char *name, char *type, void *value ) {
	if ( !dice_is_server() && dice_data.original_only == 1 && 
			dice_data.id != 0 )
		return -3;
	return dice_sync_new(name, type, value, CREATED);
}

#undef FUNCTION
#define FUNCTION		"DICE_sync_delete"
/**
  \brief Internal: Deletes an existing synchronized variable.
 
  Deletes an existing synchronized variable. Takes care of memory deallocation.

  \param name The variable name.
  \retval 0 OK.
  \retval -1 error.
  \retval -2 Variable does not exist.
  \retval -3 You do not have permission to delete variables.
  \sa DICE_sync_new.
*/
int DICE_sync_delete ( char *name ) {
	struct dice_sync		*ds;
	struct dice_sync_var_item	*item;

	_dice_debug(3, dice_printf(FUNCTION, "(%s)\n", name);)

	if ( dice_data.original_only == 1 && dice_data.id != 0 )
		return -3;

	if ( !name ) {
		_dice_debug(2, dice_printf(FUNCTION, "NULL name.\n");)
		return -1;
	}
	ds = (struct dice_sync *)hash_data(&dice_sync_var, name);
	if ( !ds ) {
		_dice_debug(2, dice_printf(FUNCTION, "variable not found.\n");)
		return -2;
	}

	/* check hierarchy */
	if ( dice_is_server() ) {
		switch ( dice_control_hierarchy(ds->changedby, packetfrom) ) {
		  case 0: /* not superior */
			return -3;
		  case 1: /* superior */
			ds->changedby = packetfrom;
			break;
		  case -1:
			  /* error*/
			break;
		}
	}

	/* check if already set for deletion */
	for ( item = sync_head; item != NULL; item = item->next )
		if ( strcmp(item->ds->name, name) == 0 ) {
			_dice_debug(2, dice_printf(FUNCTION, 
					"variable %s already deleted.\n",
					item->ds->name);)
			return 0;
		}

	ds->state = DELETED;
	dice_sync_add_to_list(ds);

	return 0;
}

#undef FUNCTION
#define FUNCTION		"DICE_sync_update"
/**
  \brief Update the content of a variable.
 
  Long description

  \param name The variable name.
  \param value The new value
  \sa DICE_sync_create.
  \retval 0 OK
  \retval -1 error.
  \retval -2 Variable doesn't exist.
  \retval -3 You do not have permission to update variables.
*/
int DICE_sync_update ( char *name, void *value ) {
	struct dice_sync		*ds;
	struct dice_sync_var_item	*item;

	_dice_debug(3, dice_printf(FUNCTION,"(%s,%p)\n",
			name, value);)

	if ( dice_data.original_only == 1 && dice_data.id != 0 )
		return -3;

	if ( !name ) {
		_dice_debug(1, dice_printf(FUNCTION, "NULL name.\n");)
		return -1;
	}
	ds = hash_data(&dice_sync_var, name);
	if ( !ds ) {
		if ( !dice_is_server() )
			_dice_debug(1, dice_printf(FUNCTION, 
				"Data %s doesn't exist.\n", name);)
		return -2;
	}

	/* check hierarchy */
	if ( dice_is_server() ) {
_dice_debug(3, dice_printf(FUNCTION,"Owned by %d, changed by %d, from %d\n",
		ds->owner, ds->changedby, packetfrom);)
		switch ( dice_control_hierarchy(ds->changedby, packetfrom) ) {
		  case 0: 
			_dice_debug(3, dice_printf(FUNCTION,"Less hierarchy\n");)
			return -3;
		  case 1: /* superior */
			ds->changedby = packetfrom;
			break;
		  case -1:
			  /* error*/
			break;
		}
	}

	ds->nextvalue = value;
	ds->state = CHANGED;

	/* avoid sending two updates for the same variable */
	for ( item = sync_head; item != NULL; item = item->next )
		if ( strcmp(item->ds->name, name) == 0 )
			return 0;

	dice_sync_add_to_list(ds);

	return 0;
}

#undef FUNCTION
#define FUNCTION		"DICE_sync"
/**
  \brief Synchronizes the applications.
 
  What it does:
  
  - Send SYNC packet
  - Send VARIABLE_SYNC packets
  - Send READY
  - Wait for packets: get VARIABLE_SYNCs, get SYNC 
  - Send READY
  - Wait for CONTINUE packet 

  \retval 0 OK
  \retval -1 error.
*/
int DICE_sync ( void ) {
	dice_packet			packet;
#ifdef USE_ASYNC
	struct sigaction		action, ignore;
	int				flags;
#endif

	_dice_debug(3, dice_printf(FUNCTION, "()\n");)

#ifdef USE_ASYNC
	/* ignore all SIGIO while we are syncing */
	ignore.sa_handler = SIG_IGN;
	sigemptyset(&ignore.sa_mask);
	sigaction(SIGIO, &ignore, &action);
	flags = fcntl(dice_data.controlsock, F_GETFL, 0);
	if ( flags == -1 ) {
		/*error */
	}
	fcntl(dice_data.controlsock, F_SETFL, flags &(~O_ASYNC));
#endif

	/* send SYNC */
	packet.packtype = SYNC;
	dice_packet_send(&packet, dice_data.controlsock);

	/* async variables may have been sent meanwhile 
	packet = dice_packet_get(dice_data.controlsock);
	while ( packet.packtype == VARIABLE_ASYNC ) {
		DICE_async_update(packet.variable.name, packet.variable.value);
		packet = dice_packet_get(dice_data.controlsock);
	} */

	/* send VARIABLEs, send READY */
	dice_sync_update_send();
	packet.packtype = READY;
	dice_packet_send(&packet, dice_data.controlsock);

	/* wait for packets; get VARIABLEs, get SYNC */
	packet = dice_packet_get(dice_data.controlsock);
	while ( packet.packtype == VARIABLE_SYNC ||
		packet.packtype == VARIABLE_FREE_SYNC ) {
		if ( packet.packtype == VARIABLE_SYNC )
			dice_sync_update(packet.variable.name, packet.variable.type,
				packet.variable.value);
/* todo: free memory here */
		else
			dice_sync_delete(packet.variable.name);
		packet = dice_packet_get(dice_data.controlsock);
	}
	if ( packet.packtype != SYNC ) {
#ifndef USE_ASYNC
		if ( packet.packtype == GOODBYE ) {
			_dice_debug(1, dice_printf(FUNCTION, 
				"got a goodbye, quitting.\n");)
			exit(0);
		}
#endif
		_dice_debug(1, dice_printf(FUNCTION, 
				"expecting SYNC, got %s\n", 
				dice_packet_name(packet.packtype));)
#ifdef USE_ASYNC
		sigaction(SIGIO, &action, NULL);
		fcntl(dice_data.controlsock, F_SETFL, flags);
#endif
		return -1;
	}

	/* send READY */
	packet.packtype = READY;
	dice_packet_send(&packet, dice_data.controlsock);

	/* wait CONTINUE */
#ifdef tcpipsync
	packet = dice_packet_get(dice_data.controlsock);
	if ( packet.packtype != CONTINUE ) {
		_dice_debug(1, dice_printf(FUNCTION,
				"expecting CONTINUE, got %s", 
				dice_packet_name(packet.packtype));)
#ifdef USE_ASYNC
		sigaction(SIGIO, &action, NULL);
		fcntl(dice_data.controlsock, F_SETFL, flags);
#endif
		return -1;
	}
#else
	dice_multicast_get();
#endif


#ifdef USE_ASYNC
	/* restore signal handler */
	sigaction(SIGIO, &action, NULL);
	fcntl(dice_data.controlsock, F_SETFL, flags);
#endif
	return 0;
}

#undef FUNCTION
#define FUNCTION 		"DICE_sync_get"
/**
  \brief Get the content of a variable.
 
  Do not free the variable value, and do not change it directly. You should
  always work on a copy of it.

  \param name The variable name.
  \sa DICE_sync_create.
  \return Variable content (cast to void *), or NULL if error.
*/
void *DICE_sync_get ( char *name ) {
	struct dice_sync		*ds;

	_dice_debug(3, dice_printf(FUNCTION, "(%s)\n",
			name);)

	if ( !name ) {
		_dice_debug(1, dice_printf(FUNCTION, "NULL name.\n");)
		return NULL;
	}
	ds = hash_data(&dice_sync_var, name);
	if ( !ds ) {
		_dice_debug(1, dice_printf(FUNCTION, "variable not found.\n");)
		return NULL;
	}
	return ds->value;
}

#undef FUNCTION
#define FUNCTION		"DICE_sync_get_next"
/**
  \brief Get the content of the next update.
 
  Long description

  \param name The variable name.
  \sa DICE_sync_create.
  \return Old variable content(cast to void *), or NULL if error.
*/
void *DICE_sync_get_next ( char *name ) {
	struct dice_sync		*ds;

	_dice_debug(3, dice_printf("DICE_sync_get_next(%s)\n",
			name);)

	if ( !name ) {
		_dice_debug(1, dice_printf(FUNCTION, "NULL name.\n");)
		return NULL;
	}
	ds = hash_data(&dice_sync_var, name);
	if ( !ds ) {
		_dice_debug(1, dice_printf(FUNCTION, "variable not found.\n");)
		return NULL;
	}

	return ds->nextvalue;
}


#undef FUNCTION
#define FUNCTION		"DICE_usync"
/**
  \brief Get the content of the next update.
 
  Long description

  \param name The variable name.
  \sa DICE_sync_create.
  \return Old variable content(cast to void *), or NULL if error.
*/
void DICE_usync ( void ) {
	static int packet_num = 0;

	_dice_debug(3, dice_printf(FUNCTION, "DICE_usync()\n");)

	/* build packet containing node number + packet number */
	do {
		/* send to server*/
//		dice_multicast_send();
		/* wait time */
//		usleep();
	} while (0/* !dice_udp_get_packet() */);

	return;
}
